(*

Ported to Ocaml from Roman Podobedov demo.

Original Copyright:
/*==========================================================================*/
/*  OpenGL Fountain 3D Demo (Particles)                                     */
/*  Author: Roman Podobedov                                                 */
/*  Email: romka@ut.ee                                                      */
/*  WEB: http://romka.demonews.com                                          */
/*  Date: 10 September 2000                                                 */
/*  Update: 30 September 2000                                               */
/*==========================================================================*/

*)

open Gl
module B = Bigarray

let pi = 3.14159

let a = Hgl.ba_double 16 
let lookAt ex ey ez cx cy cz ux uy uz =
	let z0 = ex -. cx and z1 = ey -. cy and z2 = ez -. cz in
	let mag = sqrt (z0 *. z0 +. z1 *. z1 +. z2 *. z2) in
	let z0 = if mag = 0.0 then z0 else z0 /. mag in
	let z1 = if mag = 0.0 then z1 else z1 /. mag in
	let z2 = if mag = 0.0 then z2 else z2 /. mag in
	let x0 = uy *. z2 -. uz *. z1 in
	let x1 = (~-. ux) *. z2 +. uz *. z0 in
	let x2 = ux *. z1 -. uy *. z0 in
	let y0 = z1 *. x2 -. z2 *. x1 in
	let y1 = z2 *. x0 -. z0 *. x2 in
	let y2 = z0 *. x1 -. z1 *. x0 in
	let mag = sqrt (y0 *. y0 +. y1 *. y1 +. y2 *. y2) in
	let y0 = if mag = 0.0 then y0 else y0 /. mag in
	let y1 = if mag = 0.0 then y1 else y1 /. mag in
	let y2 = if mag = 0.0 then y2 else y2 /. mag in
	let mag = sqrt (x0 *. x0 +. x1 *. x1 +. x2 *. x2) in
	let x0 = if mag = 0.0 then x0 else x0 /. mag in
	let x1 = if mag = 0.0 then x1 else x1 /. mag in
	let x2 = if mag = 0.0 then x2 else x2 /. mag in
	let f row col v = a.{col*4+row} <- v in
	f 0 0 x0; f 0 1 x1; f 0 2 x2; f 0 3 0.0;
	f 1 0 y0; f 1 1 y1; f 1 2 y2; f 1 3 0.0;
	f 2 0 z0; f 2 1 z1; f 2 2 z2; f 2 3 0.0;
	f 3 0 0.0; f 3 1 0.0; f 3 2 0.0; f 3 3 1.0;
	glMultMatrixd a;
	glTranslated (-.ex) (-.ey) (-.ez)
	
	
type tparticle =
{
	pt:float;
	pv:float;
	pd:float;
	px:float;
	py:float;
	pz:float;
	pxd:float;
	pzd:float;
	ptype:float;
	pa:float;
}

let textures = Hgl.ba_int 3 
let cx = 200.0
let cy = 200.0
let cz = 200.0
let ux = -5.0
let uy = 5.0
let uz = -5.0
let a = ref 0.0

let (f: tparticle list list ref) = ref [[];[];[]]

let load_ppm s = 	
    try
	let f = open_in_bin s in
	let magic = input_line f in
	let s = input_line f in
	if magic = "P6" then ignore (input_line f);
	let space_place = String.index s ' ' in
	let w = int_of_string (String.sub s 0 space_place) in
	let h = int_of_string (
	    String.sub s (space_place + 1) ((String.length s) - space_place - 1)
	) in
	let size = w * h * 3 in
	let buf = Hgl.ba_byte (w*h*3) in
	for i = 1 to h do
	    let x = (h - i) * w * 3 in
	    for j = 0 to w*3 - 1 do
		buf.{x+j} <- int_of_char (input_char f);
	    done;
	done;
	close_in f;
	(w,h,buf)
    with _ -> prerr_endline "Failed load_ppm"; exit 1

let load_texture name index = 
	let (texw,texh,teximage) = load_ppm name in
	glBindTexture cgl_texture_2d textures.{index};
	glPixelStorei cgl_unpack_alignment 1; 
	glTexImage2D 
		cgl_texture_2d 0 cgl_luminance texw texh 0 
		cgl_rgb cgl_unsigned_byte teximage;
	glTexParameteri cgl_texture_2d cgl_texture_min_filter cgl_linear;
	glTexParameteri cgl_texture_2d cgl_texture_mag_filter 0x2600;
	
	if index != 1 then glTexEnvi cgl_texture_env cgl_texture_env_mode 0x2100
	else begin
		glTexParameteri cgl_texture_2d cgl_texture_wrap_s 0x2901;
		glTexParameteri cgl_texture_2d cgl_texture_wrap_t 0x2901;
	end;
	()

let init () =
	glClearColor 0.0 0.0 0.0 0.0;
	glBlendFunc cgl_src_alpha cgl_one;
	glEnable cgl_blend;
	glEnable cgl_texture_2d;
	glGenTextures 3 textures;
	load_texture "./particle.pnm" 0;
	load_texture "./ground1.pnm" 1;
	load_texture "./ground2.pnm" 2;
	()
	
let reshape w h =
	glViewport 0 0 w h;
	glMatrixMode cgl_projection;
	glLoadIdentity ();
	glFrustum (-0.5) 0.5 (-0.5) 0.5 (1.0) 1000.0;
	glMatrixMode cgl_modelview;
	()

let new_particle () =
	let d = Random.float 4.0 -. 2.0 in
	let v = Random.float 2.0 +. 1.0 in
	{
		pt = -9.9;
		pv = v;
		pd = d;
		px = 0.0; py = 0.0; pz = 0.0;
		pxd = cos((d *. pi) /. 180.0) *. v /. 4.0;
		pzd = sin((d *. pi) /. 180.0) *. v ;
		ptype = 0.0;
		pa = 1.0;
	}

let iter_particles () =
	let dalpha = 2.0 /. float (List.length !f) *. pi in
	let d1 = !a *. pi /. 180.0 in
	let rec gg i = function
	| [] -> []
	| l::ll ->
		(* let alpha = ((float i) *. dalpha +. !a) *. pi /. 180.0 in *)
		let alpha = (float i) *. dalpha +. d1 in
		let sinalpha = sin alpha in
		let cosalpha = cos alpha in
		let draw p =
			let ttx = p.px *. cosalpha -. p.pz *. sinalpha in
			let ttz = p.px *. sinalpha +. p.pz *. cosalpha in
			let x0 = ttx -. cx and y0 = p.py -. cy and z0 = ttz -. cz in
			let x1 = y0 *. uz -. z0 *. uy in
			let y1 = z0 *. ux -. x0 *. uz in
			let z1 = x0 *. uy -. y0 *. ux in
			let d = sqrt (x1 *. x1 +. y1 *. y1 +. z1 *. z1) /. 5.0 in
			let x2 = x1 /. d in
			let z2 = z1 /. d in
				
			let rx0 = ttx -. x2 in
			let rx1 = ttx +. x2 in
			let ry0 = p.py -. uy in
			let ry1 = p.py +. uy in
			let rz0 = ttz -. z2 in
			let rz1 = ttz +. z2 in
			let _ = glColor4d 0.5 0.5 1.0 p.pa in
			let _ = glTexCoord2d 0.0 0.0 in
			let _ = glVertex3d rx0 ry0 rz0 in
			let _ = glTexCoord2d 1.0 0.0 in
			let _ = glVertex3d rx1 ry0 rz1 in
			let _ = glTexCoord2d 1.0 1.0 in
			let _ = glVertex3d rx1 ry1 rz1 in
			let _ = glTexCoord2d 0.0 1.0 in
			let _ = glVertex3d rx0 ry1 rz0 in
			()
		in
		let rec g = function
		| [] -> [new_particle (); new_particle ()]
		| p::l ->
			if p.ptype = 1.0 && p.pa <= 0.0 then g l 
			else begin 			
				let y = -.(9.8 *. p.pt *. p.pt /. 4.0) /. 2.0 +. 122.5 in
				let p =
					if p.ptype = 0.0 then
					{ p with
						px = p.px +. p.pxd;
						pz = p.pz +. p.pzd;
						pt = p.pt +. 0.1;
						py = y;
						ptype = if y < 0.0 then 1.0 else 0.0;
					} 
					else
					{ p with
						py = y;
						pa = p.pa -. 0.1;
					}
				in
				draw p;
				p :: g l
			end
		in
		g l :: gg (i+1) ll
	in
	glBindTexture cgl_texture_2d textures.{0};
	glPushMatrix ();
	glBegin cgl_quads;
	f:=gg 0 !f;
	glEnd ();
	glPopMatrix ();
	()

let display () =
	glClear cgl_color_buffer_bit;
	glLoadIdentity ();
	glBindTexture cgl_texture_2d textures.{1};
	a := !a +. 0.2; 
	lookAt cx cy cz 0.0 0.0 0.0 ux uy uz;
	glColor3d 0.9 0.9 1.0;
	glPushMatrix ();
	glRotated !a 0.0 (-1.0) 0.0;
	glBegin cgl_quads;
		glTexCoord2d 0.0 0.0; glVertex3d (-100.0) (0.0) (-100.0);
		glTexCoord2d 2.0 0.0; glVertex3d (-100.0) (0.0) (100.0);
		glTexCoord2d 2.0 2.0; glVertex3d (100.0) (0.0) (100.0);
		glTexCoord2d 0.0 2.0; glVertex3d (100.0) (0.0) (-100.0);
	glEnd ();
	glPopMatrix ();
	
	iter_particles (); 
	
	glBindTexture cgl_texture_2d textures.{2};
	glColor4d 1.0 1.0 0.0 0.8;
	glPushMatrix ();
	glRotated !a 0.0 (-1.0) 0.0;
	glBegin cgl_quads;
		glTexCoord2d 0.0 0.0; glVertex3d (-100.0) (0.0) (-100.0);
		glTexCoord2d 1.0 0.0; glVertex3d (-100.0) (0.0) (100.0);
		glTexCoord2d 1.0 1.0; glVertex3d (100.0) (0.0) (100.0);
		glTexCoord2d 0.0 1.0; glVertex3d (100.0) (0.0) (-100.0);
	glEnd ();
	glPopMatrix ();

	glFlush (); 
	Glfw.swap_buffers ();
	()
	
let kbd a b = if b then match a with
| _ when a = Glfw.key_esc -> raise Exit 
| _ when a = Glfw.key_tab ->
	let rec g v = function
	| 0 -> v
	| n -> new_particle () :: g v (n - 1)
	in
	f := List.map (fun l -> g l 500) !f;
	()
| _ -> ()

let iters = ref 0
let t0 = ref 0.0
let idle () = 
	incr iters;
	if !iters = 2000 then begin
		let t = Unix.gettimeofday () in
		Printf.printf "time: %g\n" (t -. !t0);
		raise Exit
	end

let _ =
    try
	if not (Glfw.init ()) then failwith "Glfw.init failed";
	if not (Glfw.open_window 800 600 0 0 0 8 32 0 Glfw.FULLSCREEN) then failwith "Glfw.open_window failed";
	let _ =
	    Printf.printf "alpha: %i depth: %i\n" 
		(Glfw.get_window_param Glfw.ALPHA_BITS)
		(Glfw.get_window_param Glfw.DEPTH_BITS);
	    flush stdout
	in
	t0 := Unix.gettimeofday ();
	init ();
	reshape 800 600;
	Glfw.set_window_size_cb reshape;
	Glfw.set_key_cb kbd;
	while true do
	    display ();
	    idle ();
	    if Glfw.get_window_param Glfw.OPENED = 0 then raise Exit
	done
    with 
    | Failure s -> prerr_endline ("Failure: "^s)
    | Exit -> Glfw.terminate ()

   