(*
 * 3-D gear wheels.  This program is in the public domain.
 *
 * Brian Paul
 * LablGL version by Jacques Garrigue
 * Converted to CamlGL by Nickolay Kolchin
 *
 *)
 
open Gl

let pi = acos (-1.)

(*
 * Draw a gear wheel.  You'll probably want to call this function when
 * building a display list since we do a lot of trig here.
 *
 * Input:  inner_radius - radius of hole at center
 *         outer_radius - radius at center of teeth
 *         width - width of gear
 *         teeth - number of teeth
 *         tooth_depth - depth of tooth
 *)
let gear ~inner ~outer ~width ~teeth ~tooth_depth =
  let r0 = inner
  and r1 = outer -. tooth_depth /. 2.0
  and r2 = outer +. tooth_depth /. 2.0 in

  let ta = 2.0 *. pi /. float teeth in
  let da = ta /. 4.0 in

  glShadeModel cgl_flat;
  glNormal3f 0.0 0.0 1.0;
  
  let vertex ~r ~z ?(s=0) i =
    let angle = float i *. ta +. float s *. da in
    glVertex3f (r *. cos angle) (r *. sin angle) z
  in

  (* draw front face *)
  let z = width *. 0.5 in
  glBegin cgl_quad_strip;
  for i=0 to teeth do
    vertex i ~r:r0 ~z;
    vertex i ~r:r1 ~z;
    vertex i ~r:r0 ~z;
    vertex i ~r:r1 ~z ~s:3;
  done;
  glEnd ();
  
  (* draw front sides of teeth *)
  glBegin cgl_quads;
  for i=0 to teeth - 1 do
    vertex i ~r:r1 ~z;
    vertex i ~r:r2 ~s:1 ~z;
    vertex i ~r:r2 ~s:2 ~z;
    vertex i ~r:r1 ~s:3 ~z;
  done;
  glEnd ();

  glNormal3f 0.0 0.0 (-1.0);

  (* draw back face *)
  let z = -. width *. 0.5 in
  glBegin cgl_quad_strip;
  for i=0 to teeth do
    vertex i ~r:r1 ~z;
    vertex i ~r:r0 ~z;
    vertex i ~r:r1 ~s:3 ~z;
    vertex i ~r:r0 ~z;
  done;
  glEnd ();

  (* draw back sides of teeth *)
  glBegin cgl_quads;
  for i=0 to teeth - 1 do
    vertex i ~r:r1 ~s:3 ~z;
    vertex i ~r:r2 ~s:2 ~z;
    vertex i ~r:r2 ~s:1 ~z;
    vertex i ~r:r1 ~z;
  done;
  glEnd ();

  (* draw outward faces of teeth *)
  let z = width *. 0.5 and z' = width *. (-0.5) in
  glBegin cgl_quad_strip;
  for i=0 to teeth - 1 do
    let angle = float i *. ta in
    vertex i ~r:r1 ~z;
    vertex i ~r:r1 ~z:z';
    let u = r2 *. cos(angle+.da) -. r1 *. cos(angle)
    and v = r2 *. sin(angle+.da) -. r1 *. sin(angle) in
    glNormal3f v (~-.u) 0.0;
    vertex i ~r:r2 ~s:1 ~z;
    vertex i ~r:r2 ~s:1 ~z:z';
    glNormal3f (cos angle) (sin angle) 0.0;
    vertex i ~r:r2 ~s:2 ~z;
    vertex i ~r:r2 ~s:2 ~z:z';
    let u = r1 *. cos(angle +. 3. *. da) -. r2 *. cos(angle +. 2. *. da)
    and v = r1 *. sin(angle +. 3. *. da) -. r2 *. sin(angle +. 2. *. da) in
    glNormal3f v (~-.u) 0.0;
    vertex i ~r:r1 ~s:3 ~z;
    vertex i ~r:r1 ~s:3 ~z:z';
    glNormal3f (cos angle) (sin angle) 0.0;
  done;
  vertex 0 ~r:r1 ~z;
  vertex 0 ~r:r1 ~z:z';
  glEnd ();

  glShadeModel cgl_smooth;

  (* draw inside radius cylinder *)
  glBegin cgl_quad_strip;
  for i=0 to teeth do
    let angle = float i *. ta in
    glNormal3f (-. cos angle) (-. sin angle) 0.0;
    vertex i ~r:r0 ~z:z';
    vertex i ~r:r0 ~z;
  done;
  glEnd ()

class view ~gear1 ~gear2 ~gear3 ?(limit=0) () = object (self)
  val mutable view_rotx = 0.0
  val mutable view_roty = 45.0
  val mutable view_rotz = 0.0
  val mutable angle = 0.0
  val mutable count = 1

  method rotx a = view_rotx <- a
  method roty a = view_roty <- a

  method draw  () =
    glClear (cgl_color_buffer_bit lor cgl_depth_buffer_bit);
    glPushMatrix ();
    glRotatef view_rotx 1.0 0.0 0.0;
    glRotatef view_roty 0.0 1.0 0.0;
    glRotatef view_rotz 0.0 0.0 1.0;

    glPushMatrix ();
    glTranslatef (-3.0) (-2.0) 0.0;
    glRotatef angle 0.0 0.0 1.0;
    
    (* gear inner:1.0 outer:4.0 width:1.0 teeth:20 tooth_depth:0.7; *)
    glCallList gear1;
    glPopMatrix ();

    glPushMatrix ();
    glTranslatef 3.1 (-2.0) 0.0;
    glRotatef (-2.0 *. angle -. 9.0) 0.0 0.0 1.0;
    (* gear inner:0.5 outer:2.0 width:2.0 teeth:10 tooth_depth:0.7; *)
    glCallList gear2;
    glPopMatrix ();

    glPushMatrix ();
    glTranslatef (-3.1) 4.2 0.0;
    glRotatef (-2.0 *. angle -. 25.0) 0.0 0.0 1.0;
    (* gear inner:1.3 outer:2.0 width:0.5 teeth:10 tooth_depth:0.7; *)
    glCallList gear3;
    glPopMatrix ();
    
    glPopMatrix ();

    Glfw.swap_buffers ();

    count <- count + 1;
    if count = limit then exit 0

  method idle () =
    angle <- angle +. 0.2;
    self#draw ()

  method reshape w h =
    glViewport 0 0 w h;
    glMatrixMode cgl_projection;
    glLoadIdentity ();
    let r = float w /. float h in
    let r' = 1. /. r in
    if (w>h) then
      glFrustum (~-.r) r (-1.0) 1.0 5.0 60.0
    else
      glFrustum (-1.0) 1.0 (~-.r') r' 5.0 60.0;
    glMatrixMode cgl_modelview;
    glLoadIdentity ();
    glTranslatef 0.0 0.0 (-40.0);
    glClear (cgl_color_buffer_bit lor cgl_depth_buffer_bit)
end

let init () =
  let pos = Hgl.ba_float_init [|5.0;5.0;10.0;0.0|]
  and red = Hgl.ba_float_init [|0.8;0.1;0.0;1.0|]
  and green = Hgl.ba_float_init [|0.0;0.8;0.2;1.0|]
  and blue = Hgl.ba_float_init [|0.2;0.2;1.0;1.0|] in

  glLightfv cgl_light0 cgl_position pos;
  List.iter glEnable
    [cgl_cull_face;cgl_lighting;cgl_light0;cgl_depth_test;cgl_normalize];

  (* make the gears *)
  let make_gear ~inner ~outer ~width ~teeth ~color =
    Hgl.make_list (fun () ->
	glMaterialfv cgl_front cgl_ambient_and_diffuse color;
	gear ~inner ~outer ~width ~teeth ~tooth_depth:0.7;
    )
  in
  let gear1 = make_gear ~inner:1.0 ~outer:4.0 ~width:1.0 ~teeth:20 ~color:red
  and gear2 = make_gear ~inner:0.5 ~outer:2.0 ~width:2.0 ~teeth:10 ~color:green
  and gear3 = make_gear ~inner:1.3 ~outer:2.0 ~width:0.5 ~teeth:10 ~color:blue in

  (gear1, gear2, gear3)

let main () =
    try
	if not (Glfw.init ()) then failwith "Glfw.init failed";
	if not (Glfw.open_window 800 600 0 0 0 8 24 0 Glfw.WINDOW) then failwith "Glfw.open_window failed";
	Glfw.set_window_title "gears";
	let gear1, gear2, gear3 = init() in
        let view = new view ~gear1 ~gear2 ~gear3 () in
	Glfw.enable Glfw.STICKY_KEYS;
	Glfw.swap_interval 1;
	view#reshape 800 600;
	while true do
	    view#draw ();
	    view#idle ();
    	    if Glfw.get_window_param Glfw.OPENED = 0 || Glfw.get_key Glfw.key_esc then 
		raise Exit;
	done
    with
    | Failure s -> Printf.eprintf "Failure: %s\n" s
    | Exit -> Glfw.terminate ()

let _ = main ()
